/*******************************************
 * Copyright(c) 2000-2003 Mikio Hirabayashi
 *******************************************/


#include "curia.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#define CR_DIRMODE     00755             /* ǥ쥯ȥΥ⡼ */
#define CR_FILEMODE    00644             /* եΥ⡼ */
#define CR_PATHBUFSIZ  1024              /* ѥѤΥХåե */
#define CR_NUMBUFSIZ   32                /* ͤʸɽΥХåե */
#define CR_DEFDNUM     5                 /* ǥեȤǥǡ١ */
#define CR_DPNAME      "depot"           /* ǡ١ե̾ */
#define CR_KEYDNUM     "dnum"            /* ʬ° */
#define CR_KEYLRNUM    "lrnum"           /* 顼֥ȿ° */
#define CR_LOBDIR      "lob"             /* 顼֥ѥǥ쥯ȥ̾ */
#define CR_LOBDLEVEL   3                 /* 顼֥ѥǥ쥯ȥγؿ */


static int crdpgetnum(DEPOT *depot, const char *kbuf, int ksiz);
static char *crgetlobpath(CURIA *curia, const char *kbuf, int ksiz);
static int crmklobdir(const char *path);
static int crrmlobdir(const char *path);
static int crwrite(int fd, const void *buf, int size);
static int crread(int fd, void *buf, int size);



/*************************************************************************************************
 * ֥
 *************************************************************************************************/


/* ǡ١ǥ쥯ȥ򳫤Υϥɥ֤ */
CURIA *cropen(const char *name, CROMODE omode, int bnum, int dnum){
  DEPOT *attr, **depots;
  DPOMODE dpomode;
  char path[CR_PATHBUFSIZ], *tname;
  int i, j, lrnum;
  CURIA *curia;
  assert(name);
  /* ѥ᡼ */
  if(dnum < 1 || dnum >= CR_DPMAX) dnum = CR_DEFDNUM;
  if(strlen(name) > CR_NAMEMAX){
    dpecode = DP_EMISC;
    return NULL;
  }
  dpomode = DP_OREADER;
  if(omode & CR_OWRITER){
    dpomode = DP_OWRITER;
    if(omode & CR_OCREAT) dpomode |= DP_OCREAT;
    if(omode & CR_OTRUNC) dpomode |= DP_OTRUNC;
  }
  attr = NULL;
  lrnum = 0;
  /* ν */
  if((omode & CR_OWRITER) && (omode & CR_OCREAT)){
    /* ١ǥ쥯ȥκ */
    if(mkdir(name, CR_DIRMODE) == -1 && errno != EEXIST){
      dpecode = DP_EMKDIR;
      return NULL;
    }
    /* ǥǥ쥯ȥκ */
    for(i = 0; i < dnum; i++){
      sprintf(path, "%s/%04d", name, i + 1);
      if(mkdir(path, CR_DIRMODE) == -1 && errno != EEXIST){
        dpecode = DP_EMKDIR;
        return NULL;
      }
    }
    /* °ǡ١κ */
    sprintf(path, "%s/%s", name, CR_DPNAME);
    if(!(attr = dpopen(path, dpomode, 1))) return NULL;
    if(!dpput(attr, CR_KEYDNUM, -1, (char *)&dnum, sizeof(int), DP_DOVER)){
      dpclose(attr);
      return NULL;
    }
  }
  /* °ǡ١Υץ°ͤμ */
  if(!attr){
    sprintf(path, "%s/%s", name, CR_DPNAME);
    if(!(attr = dpopen(path, dpomode, 1))) return NULL;
    if(!(omode & CR_OTRUNC)){
      if((dnum = crdpgetnum(attr, CR_KEYDNUM, -1)) < 1 ||
         (lrnum = crdpgetnum(attr, CR_KEYLRNUM, -1)) < 0){
        dpclose(attr);
        dpecode = DP_EBROKEN;
        return NULL;
      }
    }
  }
  /* ڤͤ */
  if(omode & CR_OTRUNC){
    for(i = 0; i < CR_DPMAX; i++){
      /* ǡ١κ */
      sprintf(path, "%s/%04d/%s", name, i + 1, CR_DPNAME);
      if(unlink(path) == -1 && errno != ENOENT){
        dpclose(attr);
        dpecode = DP_EUNLINK;
        return NULL;
      }
      /* 顼֥Ȥκ */
      sprintf(path, "%s/%04d/%s", name, i + 1, CR_LOBDIR);
      if(!crrmlobdir(path)){
        dpclose(attr);
        return NULL;
      }
      /* ֥ǥ쥯ȥκ */
      if(i >= dnum){
        sprintf(path, "%s/%04d", name, i + 1);
        if(rmdir(path) == -1 && errno != ENOENT){
          dpclose(attr);
          dpecode = DP_ERMDIR;
          return NULL;
        }
      }
    }
    errno = 0;
  }
  /* ǥǡ١Υץ */
  if(!(depots = malloc(dnum * sizeof(DEPOT *)))){
    dpclose(attr);
    dpecode = DP_EALLOC;
    return NULL;
  }
  for(i = 0; i < dnum; i++){
    sprintf(path, "%s/%04d/%s", name, i + 1, CR_DPNAME);
    if(!(depots[i] = dpopen(path, dpomode, bnum))){
      for(j = 0; j < i; j++){
        dpclose(depots[j]);
      }
      free(depots);
      dpclose(attr);
      return NULL;
    }
  }
  /* ϥɥ빽¤Τ */
  curia = malloc(sizeof(CURIA));
  tname = strdup(name);
  if(!curia || !tname){
    free(curia);
    free(tname);
    for(i = 0; i < dnum; i++){
      dpclose(depots[i]);
    }
    free(depots);
    dpclose(attr);
    dpecode = DP_EALLOC;
    return NULL;
  }
  curia->name = tname;
  curia->wmode = (omode & CR_OWRITER);
  curia->attr = attr;
  curia->depots = depots;
  curia->dnum = dnum;
  curia->inum = 0;
  curia->lrnum = lrnum;
  return curia;
}


/* ǡ١ǥ쥯ȥĤ롣 */
int crclose(CURIA *curia){
  int i, err;
  assert(curia);
  err = FALSE;
  /* ǥǡ١Υ */
  for(i = 0; i < curia->dnum; i++){
    if(!dpclose(curia->depots[i])) err = TRUE;
  }
  free(curia->depots);
  /* °ȿ */
  if(curia->wmode){
    if(!dpput(curia->attr, CR_KEYLRNUM, -1, (char *)&(curia->lrnum), sizeof(int), DP_DOVER))
      err = TRUE;
  }
  /* °ǡ١Υ */
  if(!dpclose(curia->attr)) err = TRUE;
  /* ¾Υ꥽ */
  free(curia->name);
  free(curia);
  return err ? FALSE : TRUE;
}


/* ǡ١˥쥳ɤɲä롣 */
int crput(CURIA *curia, const char *kbuf, int ksiz, const char *vbuf, int vsiz, CRDMODE dmode){
  DPDMODE dpdmode;
  int tnum;
  assert(curia && kbuf && vbuf);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ⡼ɤɤؤ */
  switch(dmode){
  case CR_DKEEP: dpdmode = DP_DKEEP; break;
  case CR_DCAT: dpdmode = DP_DCAT; break;
  default: dpdmode = DP_DOVER; break;
  }
  /* ǥǡ١ */
  tnum = dpouterhash(kbuf, ksiz) % curia->dnum;
  return dpput(curia->depots[tnum], kbuf, ksiz, vbuf, vsiz, dpdmode);
}


/* ǡ١쥳ɤ롣 */
int crout(CURIA *curia, const char *kbuf, int ksiz){
  int tnum;
  assert(curia && kbuf);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ǥǡ١ */
  tnum = dpouterhash(kbuf, ksiz) % curia->dnum;
  return dpout(curia->depots[tnum], kbuf, ksiz);
}


/* ǡ١쥳ɤͤФ롣 */
char *crget(CURIA *curia, const char *kbuf, int ksiz, int start, int max, int *sp){
  int tnum;
  assert(curia && kbuf && start >= 0);
  /* ǥǡ١ */
  tnum = dpouterhash(kbuf, ksiz) % curia->dnum;
  return dpget(curia->depots[tnum], kbuf, ksiz, start, max, sp);
}


/* ǡ١ˤ쥳ɤͤΥ롣 */
int crvsiz(CURIA *curia, const char *kbuf, int ksiz){
  int tnum;
  assert(curia && kbuf);
  /* ǥǡ١ */
  tnum = dpouterhash(kbuf, ksiz) % curia->dnum;
  return dpvsiz(curia->depots[tnum], kbuf, ksiz);
}


/* ǡ١ΥΥƥ졼롣 */
int criterinit(CURIA *curia){
  int i, err;
  assert(curia);
  /* ǥǡ١ν */
  err = FALSE;
  for(i = 0; i < curia->dnum; i++){
    if(!dpiterinit(curia->depots[i])){
      err = TRUE;
      break;
    }
  }
  /* ǥåν */
  curia->inum = 0;
  return err ? FALSE : TRUE;
}


/* ǡ١ΥΥƥ졼鼡ΥФ */
char *criternext(CURIA *curia, int *sp){
  char *vbuf;
  assert(curia);
  vbuf = NULL;
  /* ǥǡ١Υƥ졼˻ */
  while(curia->inum < curia->dnum && !(vbuf = dpiternext(curia->depots[curia->inum], sp))){
    if(dpecode != DP_ENOITEM) return NULL;
    (curia->inum)++;
  }
  return vbuf;
}


/* ǡ١Υ饤Ȥꤹ롣 */
int crsetalign(CURIA *curia, int asiz, int aunit){
  int i, err;
  assert(curia && asiz > 0 && aunit >= 0);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ǥǡ١Υ饤 */
  err = FALSE;
  for(i = 0; i < curia->dnum; i++){
    if(!dpsetalign(curia->depots[i], asiz, aunit)){
      err = TRUE;
      break;
    }
  }
  return err ? FALSE : TRUE;
}


/* ǡ١Ƥե뤪ӥǥХƱ롣 */
int crsync(CURIA *curia){
  int i, err;
  assert(curia);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  err = FALSE;
  /* ǥǡ١Ʊ */
  for(i = 0; i < curia->dnum; i++){
    if(!dpsync(curia->depots[i])){
      err = TRUE;
      break;
    }
  }
  return err ? FALSE : TRUE;  
}


/* ǡ١Ŭ롣 */
int croptimize(CURIA *curia, int bnum){
  int i, err;
  assert(curia);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  err = FALSE;
  /* ǥǡ١Ʊ */
  for(i = 0; i < curia->dnum; i++){
    if(!dpoptimize(curia->depots[i], bnum)){
      err = TRUE;
      break;
    }
  }
  curia->inum = 0;
  return err ? FALSE : TRUE;
}


/* ǡ١֤̾ */
char *crname(CURIA *curia){
  char *name;
  assert(curia);
  /* ̾ʣ */
  if(!(name = strdup(curia->name))){
    dpecode = DP_EALLOC;
    return NULL;
  }
  return name;
}


/* ǥǡ١Υե륵ιפ֤ */
int crfsiz(CURIA *curia){
  int i, sum, rv;
  assert(curia);
  /* ǥǡ١ι */
  sum = 0;
  for(i = 0; i < curia->dnum; i++){
    rv = dpfsiz(curia->depots[i]);
    if(rv == -1) return -1;
    sum += rv;
  }
  return sum;
}


/* ǥǡ١ΥХåǿιפ֤ */
int crbnum(CURIA *curia){
  int i, sum, rv;
  assert(curia);
  /* ǥǡ١ι */
  sum = 0;
  for(i = 0; i < curia->dnum; i++){
    rv = dpbnum(curia->depots[i]);
    if(rv == -1) return -1;
    sum += rv;
  }
  return sum;
}


/* ǡ١ΥХåǿιפ֤ */
int crbusenum(CURIA *curia){
  int i, sum, rv;
  assert(curia);
  /* ǥǡ١ι */
  sum = 0;
  for(i = 0; i < curia->dnum; i++){
    rv = dpbusenum(curia->depots[i]);
    if(rv == -1) return -1;
    sum += rv;
  }
  return sum;
}


/* ǥǡ١Υ쥳ɿιפ֤ */
int crrnum(CURIA *curia){
  int i, sum, rv;
  assert(curia);
  /* ǥǡ١ι */
  sum = 0;
  for(i = 0; i < curia->dnum; i++){
    rv = dprnum(curia->depots[i]);
    if(rv == -1) return -1;
    sum += rv;
  }
  return sum;
}


/* 顼֥ѥǡ١˥쥳ɤɲä롣 */
int crputlob(CURIA *curia, const char *kbuf, int ksiz, const char *vbuf, int vsiz, CRDMODE dmode){
  char *path;
  int mode, fd, err, be;
  struct stat sbuf;
  assert(curia && kbuf && vbuf);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ѥ᡼ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  if(vsiz < 0) vsiz = strlen(vbuf);
  /* եѥγ */
  if(!(path = crgetlobpath(curia, kbuf, ksiz))) return FALSE;
  if(!crmklobdir(path)){
    free(path);
    return FALSE;
  }
  /* ¸եθ */
  be = stat(path, &sbuf) != -1 && S_ISREG(sbuf.st_mode);
  /* եΥץ */
  mode = O_RDWR | O_CREAT;
  if(dmode & CR_DKEEP) mode |= O_EXCL;
  if(dmode & CR_DCAT){
    mode |= O_APPEND;
  } else {
    mode |= O_TRUNC;
  }
  if((fd = open(path, mode, CR_FILEMODE)) == -1){
    free(path);
    dpecode = DP_EOPEN;
    if(dmode == CR_DKEEP) dpecode = DP_EKEEP;
    return FALSE;
  }
  free(path);
  /* ǡν񤭹 */
  err = FALSE;
  if(crwrite(fd, vbuf, vsiz) == -1){
    dpecode = DP_EWRITE;
    err = TRUE;
  }
  /* եΥ */
  if(close(fd) == -1){
    dpecode = DP_ECLOSE;
    err = TRUE;
  }
  /* 쥳ɿι */
  if(!err && !be) (curia->lrnum)++;
  return err ? FALSE : TRUE;
}


/* 顼֥ѥǡ١쥳ɤ롣 */
int croutlob(CURIA *curia, const char *kbuf, int ksiz){
  char *path;
  int err, be;
  struct stat sbuf;
  assert(curia && kbuf);
  /* ϥɥθ */
  if(!curia->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ѥ᡼ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  /* եѥλ */
  if(!(path = crgetlobpath(curia, kbuf, ksiz))) return FALSE;
  /* ¸եθ */
  be = stat(path, &sbuf) != -1 && S_ISREG(sbuf.st_mode);
  /* եκ */
  err = FALSE;
  if(unlink(path) == -1){
    dpecode = DP_ENOITEM;
    err = TRUE;
  }
  free(path);
  /* 쥳ɿι */
  if(!err && be) (curia->lrnum)--;
  return err ? FALSE : TRUE;
}


/* 顼֥ѥǡ١쥳ɤͤФ롣 */
char *crgetlob(CURIA *curia, const char *kbuf, int ksiz, int start, int max, int *sp){
  char *path, *buf;
  struct stat sbuf;
  int fd, size;
  assert(curia && kbuf && start >= 0);
  /* ѥ᡼ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  /* եѥλ */
  if(!(path = crgetlobpath(curia, kbuf, ksiz))) return NULL;
  /* եΥץ */
  if((fd = open(path, O_RDONLY, CR_FILEMODE)) == -1){
    free(path);
    dpecode = DP_ENOITEM;
    return NULL;
  }
  free(path);
  /* Ƚ */
  if(fstat(fd, &sbuf) == -1){
    close(fd);
    dpecode = DP_ESTAT;
    return NULL;
  }
  /* ɤ߹߳ޤǰư */
  if(start > sbuf.st_size){
    close(fd);
    dpecode = DP_ENOITEM;
    return NULL;
  }
  if(lseek(fd, start, SEEK_SET) == -1){
    close(fd);
    dpecode = DP_ESEEK;
    return NULL;
  }
  /* ɤ߹ */
  if(max < 0) max = sbuf.st_size;
  if(!(buf = malloc(max + 1))){
    close(fd);
    dpecode = DP_EALLOC;
    return NULL;
  }
  size = crread(fd, buf, max);
  /* եΥ */
  close(fd);
  if(size == -1){
    free(buf);
    dpecode = DP_EREAD;
    return NULL;
  }
  buf[size] = '\0';
  if(sp) *sp = size;
  return buf;
}


/* 顼֥ѥǡ١ˤ쥳ɤͤΥ롣 */
int crvsizlob(CURIA *curia, const char *kbuf, int ksiz){
  char *path;
  struct stat sbuf;
  assert(curia && kbuf);
  /* ѥ᡼ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  /* եѥλ */
  if(!(path = crgetlobpath(curia, kbuf, ksiz))) return -1;
  /* Ƚ */
  if(stat(path, &sbuf) == -1){
    free(path);
    dpecode = DP_ENOITEM;
    return -1;
  }
  free(path);
  return sbuf.st_size;
}


/* 顼֥ѥǡ١Υ쥳ɿ֤ */
int crrnumlob(CURIA *curia){
  assert(curia);
  return curia->lrnum;
}



/*************************************************************************************************
 * ֥
 *************************************************************************************************/


/* °ǡ١intοͤФ
   depot °ǡ١Υϥɥꤹ롣
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ뤫ʤ kbuf ʸĹȤʤ롣
   ͤʤФ͡顼ʤ INT_MIN Ǥ롣 */
static int crdpgetnum(DEPOT *depot, const char *kbuf, int ksiz){
  char *vbuf;
  int vsiz, rv;
  if(!(vbuf = dpget(depot, kbuf, ksiz, 0, -1, &vsiz)) || vsiz != sizeof(int)){
    free(vbuf);
    return INT_MIN;
  }
  rv = *(int *)vbuf;
  free(vbuf);
  return rv;
}


/* б顼֥ȥեΥѥ֤
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   ͤϥեѥǼȤ줿ΰΥݥ󥿤顼ʤ NULL Ǥ롣 */
static char *crgetlobpath(CURIA *curia, const char *kbuf, int ksiz){
  char prefix[CR_PATHBUFSIZ], *wp, *path;
  int i, hash;
  assert(curia && kbuf && ksiz >= 0);
  wp = prefix;
  wp += sprintf(wp, "%s/%04d/%s/",
                curia->name, dpouterhash(kbuf, ksiz) % curia->dnum + 1, CR_LOBDIR);
  hash = dpinnerhash(kbuf, ksiz);
  for(i = 0; i < CR_LOBDLEVEL; i++){
    wp += sprintf(wp, "%02X/", hash % 0x100);
    hash /= 0x100;
  }
  if(!(path = malloc(strlen(prefix) + ksiz * 2 + 1))){
    dpecode = DP_EALLOC;
    return NULL;
  }
  wp = path;
  wp += sprintf(path, "%s", prefix);
  for(i = 0; i < ksiz; i++){
    wp += sprintf(wp, "%02X", ((unsigned char *)kbuf)[i]);
  }
  return path;
}


/* ѥʬ˴ޤޤǥ쥯ȥ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣
   ˤΥǥ쥯ȥ꤬ϥ顼Ȥʤ */
static int crmklobdir(const char *path){
  char elem[CR_PATHBUFSIZ], *wp;
  const char *dp;
  int err, len;
  wp = elem;
  err = FALSE;
  while(*path != '\0' && (dp = strchr(path, '/')) != NULL){
    len = dp - path;
    if(wp != elem) wp += sprintf(wp, "/");
    memcpy(wp, path, len);
    wp[len] = '\0';
    wp += len;
    if(mkdir(elem, CR_DIRMODE) == -1 && errno != EEXIST) err = TRUE;
    path = dp + 1;
  }
  if(err) dpecode = DP_EMKDIR;
  return err ? FALSE : TRUE;
}


/* 顼֥ȤΥǥ쥯ȥƵŪ˺롣
   path Ϻǥ쥯ȥΥѥꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣
   ܥå󥯤Ϥ켫ȤäơפϤʤ */
static int crrmlobdir(const char *path){
  char elem[CR_PATHBUFSIZ];
  DIR *DD;
  struct dirent *dp;
  assert(path);
  if(unlink(path) != -1){
    return TRUE;
  } else {
    if(errno == ENOENT) return TRUE;
    if(!(DD = opendir(path))){
      dpecode = DP_EMISC;
      return FALSE;
    }
    while((dp = readdir(DD)) != NULL){
      if(!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, "..")) continue;
      sprintf(elem, "%s/%s", path, dp->d_name);
      if(!crrmlobdir(elem)){
        closedir(DD);
        return FALSE;
      }
    }
  }
  if(closedir(DD) == -1){
    dpecode = DP_EMISC;
    return FALSE;
  }
  if(rmdir(path) == -1){
    dpecode = DP_ERMDIR;
    return FALSE;
  }
  return TRUE;
}


/* ե˥ХåեΥǡƽ񤭽Ф
   fd оݤΥեǥץꤹ롣
   buf Ͻ񤭽ФХåեꤹ롣
   size Ͻ񤭽ФХȿꤹ롣
   ͤʤ񤭽ФХȿ顼ʤ -1 Ǥ롣 */
static int crwrite(int fd, const void *buf, int size){
  char *lbuf;
  int rv, wb;
  assert(fd >= 0 && buf && size >= 0);
  lbuf = (char *)buf;
  rv = 0;
  do {
    wb = write(fd, lbuf, size);
    switch(wb){
    case -1: if(errno != EINTR) return -1;
    case 0: break;
    default:
      lbuf += wb;
      size -= wb;
      rv += wb;
      break;
    }
  } while(size > 0);
  return rv;
}


/* ե뤫ХåեΥΥǡɤ߹ࡣ
   fd оݤΥեǥץꤹ롣
   buf ϥǡ¸Хåեꤹ롣
   size ɤ߹Хȿꤹ롣
   ͤʤɤ߹Хȿ顼ʤ -1 Ǥ롣
   ͤ size 꾮֤ͤäϥե뽪ụ̈롣 */
static int crread(int fd, void *buf, int size){
  char *lbuf;
  int i, bs;
  assert(fd >= 0 && buf && size >= 0);
  lbuf = buf;
  for(i = 0; i < size && (bs = read(fd, lbuf + i, size - i)) != 0; i += bs){
    if(bs == -1 && errno != EINTR) return -1;
  }
  return i;
}



/* END OF FILE */
