/*******************************************
 * Copyright(c) 2000-2003 Mikio Hirabayashi
 *******************************************/


#include "depot.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#ifndef DP_VERSION                       /* ߡС */
#define DP_VERSION     "0.0.0"
#endif

#define DP_FILEMODE    00644             /* եΥ⡼ */
#define DP_MAGICNUM    "[depot]\n\f"     /* ޥåʥС */
#define DP_HEADSIZ     48                /* إåΥ */
#define DP_FSIZOFF     16                /* ե륵ΰΥեå */
#define DP_BNUMOFF     24                /* ХåȿΰΥեå */
#define DP_RNUMOFF     32                /* Ǽ쥳ɿΰΥեå */
#define DP_NUMBUFSIZ   32                /* ͤʸɽΥХåե */
#define DP_DEFBNUM     8191              /* ǥեȤΥХåȿ */
#define DP_OPTFSUF     ".opt"            /* Ŭΰե */
#define DP_OPTBLOAD    0.35              /* ŬΥХåȤٷ */
#define DP_OPTRUNIT    256               /* ŬΥ쥳ɰưñ */


enum {                                   /* 쥳ɥإåΥǥå */
  DP_RHIFLAGS,                           /* ե饰 */
  DP_RHIHASH,                            /* 裲ϥå */
  DP_RHIKSIZ,                            /* Υ */
  DP_RHIVSIZ,                            /* ͤΥ */
  DP_RHIPSIZ,                            /* ѥǥ󥰤Υ */
  DP_RHILEFT,                            /* λҤΥեå */
  DP_RHIRIGHT,                           /* λҤΥեå */
  DP_RHNUM                               /* 쥳ɥإåιܿ */
};

enum {                                   /* 쥳ɤΥե饰 */
  DP_RECFDEL = 1 << 0                    /*  */
};


static int dplock(int fd, int ex);
static int dpwrite(int fd, const void *buf, int size);
static int dpseekwrite(int fd, int off, const void *buf, int size);
static int dpseekwritenum(int fd, int off, int num);
static int dpread(int fd, void *buf, int size);
static int dpseekread(int fd, int off, void *buf, int size);
static int dpprintf(int fd, const char *format, ...);
static void dpdbgoutdepot(const char *fname, DEPOT *depot);
static void dpdbgouthead(const char *fname, int *head);
static int dpfirsthash(const char *kbuf, int ksiz);
static int dpsecondhash(const char *kbuf, int ksiz);
static int dpthirdhash(const char *kbuf, int ksiz);
static int dpgetprime(int num);
static int dppadsize(DEPOT *depot, int vsiz);
static int dprecsize(int *head);
static int dprechead(DEPOT *depot, int off, int *head);
static char *dpreckey(DEPOT *depot, int off, int *head);
static char *dprecval(DEPOT *depot, int off, int *head, int start, int max);
static int dpkeycmp(const char *abuf, int asiz, const char *bbuf, int bsiz);
static int dprecsearch(DEPOT *depot, const char *kbuf, int ksiz, int hash,
                       int *bip, int *offp, int *entp, int *head, int delhit);
static int dprecappend(DEPOT *depot, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
                       int hash, int left, int right);
static int dprecover(DEPOT *depot, int off, int *head, const char *vbuf, int vsiz, int cat);
static int dprecdelete(DEPOT *depot, int off, int *head);



/*************************************************************************************************
 * ֥
 *************************************************************************************************/


/* ǥХå */
int dpdbgfd = -1;


/* Сʸ */
const char *dpversion = DP_VERSION;


/* ľΥ顼 */
DPECODE dpecode = DP_ENOERR;


/* 顼ɤбåʸ֤ */
const char *dperrmsg(DPECODE ecode){
  switch(ecode){
  case DP_ENOERR: return "no error";
  case DP_EFATAL: return "with fatal error";
  case DP_EMODE: return "invalid mode";
  case DP_EBROKEN: return "broken database file";
  case DP_EKEEP: return "existing record";
  case DP_ENOITEM: return "no item found";
  case DP_EALLOC: return "memory allocation error";
  case DP_EMAP: return "memory mapping error";
  case DP_EOPEN: return "open error";
  case DP_ECLOSE: return "close error";
  case DP_ETRUNC: return "trunc error";
  case DP_ESYNC: return "sync error";
  case DP_ESTAT: return "stat error";
  case DP_ESEEK: return "seek error";
  case DP_EREAD: return "read error";
  case DP_EWRITE: return "write error";
  case DP_ELOCK: return "lock error";
  case DP_ERENAME: return "rename error";
  case DP_EUNLINK: return "unlink error";
  case DP_EMKDIR: return "mkdir error";
  case DP_ERMDIR: return "rmdir error";
  case DP_EMISC: return "miscellaneous error";
  }
  return "(invalid ecode)";
}


/* ǡ١ե򳫤Υϥɥ֤ */
DEPOT *dpopen(const char *name, DPOMODE omode, int bnum){
  int mode, fd, fsiz, rnum, c, msiz;
  char hbuf[DP_HEADSIZ], *map;
  struct stat sbuf;
  DEPOT *depot;
  assert(name);
  /* ŬˤڤкΥ롼 */
  while(TRUE){
    /* եΥץ */
    mode = O_RDONLY;
    if(omode & DP_OWRITER){
      mode = O_RDWR;
      if(omode & DP_OCREAT) mode |= O_CREAT;
    }
    if((fd = open(name, mode, DP_FILEMODE)) == -1){
      dpecode = DP_EOPEN;
      return NULL;
    }
    /* եΥå */
    if(!dplock(fd, omode & DP_OWRITER)){
      close(fd);
      return NULL;
    }
    /* Υեڤͤ */
    if((omode & DP_OWRITER) && (omode & DP_OTRUNC)){
      if(ftruncate(fd, 0) == -1){
        dpecode = DP_ETRUNC;
        close(fd);
        return NULL;
      }
    }
    /* եξֳǧȥμ */
    if(fstat(fd, &sbuf) == -1 || !S_ISREG(sbuf.st_mode)){
      close(fd);
      dpecode = DP_ESTAT;
      return NULL;
    }
    fsiz = sbuf.st_size;
    /* 󥯤ڤƤľ */
    if(sbuf.st_nlink < 1){
      close(fd);
      continue;
    }
    break;
  }
  /* ν */
  if((omode & DP_OWRITER) && fsiz == 0){
    memset(hbuf, 0, DP_HEADSIZ);
    memcpy(hbuf, DP_MAGICNUM, strlen(DP_MAGICNUM));
    bnum = bnum < 1 ? DP_DEFBNUM : bnum;
    bnum = dpgetprime(bnum);
    memcpy(hbuf + DP_BNUMOFF, &bnum, sizeof(int));
    rnum = 0;
    memcpy(hbuf + DP_RNUMOFF, &rnum, sizeof(int));
    c = 0;
    fsiz = DP_HEADSIZ + bnum * sizeof(int);
    memcpy(hbuf + DP_FSIZOFF, &fsiz, sizeof(int));
    if(!dpseekwrite(fd, 0, hbuf, DP_HEADSIZ) ||
       !dpseekwrite(fd, fsiz - 1, &c, 1)){
      close(fd);
      return NULL;
    }
  }
  /* إåɤ߹ */
  if(!dpseekread(fd, 0, hbuf, DP_HEADSIZ)){
    close(fd);
    dpecode = DP_EBROKEN;
    return NULL;
  }
  /* ޥåʥФȥե륵θ */
  if(memcmp(hbuf, DP_MAGICNUM, strlen(DP_MAGICNUM)) != 0 ||
     memcmp(hbuf + DP_FSIZOFF, &fsiz, sizeof(int)) != 0){
    close(fd);
    dpecode = DP_EBROKEN;
    return NULL;
  }
  /* Хåȿȥ쥳ɿμ */
  bnum = *((int *)(hbuf + DP_BNUMOFF));
  rnum = *((int *)(hbuf + DP_RNUMOFF));
  if(bnum < 1 || rnum < 0 || fsiz < DP_HEADSIZ + bnum * sizeof(int)){
    close(fd);
    dpecode = DP_EBROKEN;
    return NULL;
  }
  /* ޥåԥ */
  msiz = DP_HEADSIZ + bnum * sizeof(int);
  map = mmap(0, msiz, PROT_READ | ((mode & DP_OWRITER) ? PROT_WRITE : 0), MAP_SHARED, fd, 0);
  if(map == MAP_FAILED){
    close(fd);
    dpecode = DP_EMAP;
    return NULL;
  }
  /* ϥɥ빽¤Τ */
  if(!(depot = malloc(sizeof(DEPOT))) || !(depot->name = strdup(name))){
    munmap(map, msiz);
    free(depot);
    close(fd);
    dpecode = DP_EALLOC;
    return NULL;
  }
  depot->wmode = (mode & DP_OWRITER);
  depot->fd = fd;
  depot->fsiz = fsiz;
  depot->map = map;
  depot->msiz = msiz;
  depot->bucket = (int *)(map + DP_HEADSIZ);
  depot->bnum = bnum;
  depot->rnum = rnum;
  depot->fatal = FALSE;
  depot->ioff = 0;
  depot->asiz = 1;
  depot->aunit = 0;
  /* ǥХå */
  dpdbgoutdepot("dpopen", depot);
  return depot;
}


/* ǡ١եĤ롣 */
int dpclose(DEPOT *depot){
  int err;
  assert(depot);
  err = FALSE;
  /* ǥХå */
  dpdbgoutdepot("dpclose", depot);
  /* إåȥХåƱ */
  if(depot->wmode){
    *((int *)(depot->map + DP_FSIZOFF)) = depot->fsiz;
    *((int *)(depot->map + DP_RNUMOFF)) = depot->rnum;
  }
  if(munmap(depot->map, depot->msiz) == -1){
    dpecode = DP_EMAP;
    err = TRUE;
  }
  /* ꥽β */
  if(close(depot->fd) == -1){
    dpecode = DP_ECLOSE;
    err = TRUE;
  }
  free(depot->name);
  free(depot);
  return !err;
}


/* ǡ١˥쥳ɤɲä롣 */
int dpput(DEPOT *depot, const char *kbuf, int ksiz, const char *vbuf, int vsiz, DPDMODE dmode){
  int head[DP_RHNUM], hash, bi, off, entoff, newoff, rsiz, nsiz, fdel;
  char *tval;
  assert(depot && kbuf && vbuf);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  if(!depot->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* 쥳ɤθ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  if(vsiz < 0) vsiz = strlen(vbuf);
  newoff = -1;
  hash = dpsecondhash(kbuf, ksiz);
  switch(dprecsearch(depot, kbuf, ksiz, hash, &bi, &off, &entoff, head, TRUE)){
  case -1:
    /* 顼 */
    depot->fatal = TRUE;
    return FALSE;
  case 0:
    /* ʣȯ */
    fdel = head[DP_RHIFLAGS] & DP_RECFDEL;
    if(dmode == DP_DKEEP && !fdel){
      dpecode = DP_EKEEP;
      return FALSE;
    }
    /* 쥳ɤΥեåԽ */
    if(fdel){
      head[DP_RHIPSIZ] += head[DP_RHIVSIZ];
      head[DP_RHIVSIZ] = 0;
    }
    /* ¸ɬץλ */
    rsiz = dprecsize(head);
    nsiz = DP_RHNUM * sizeof(int) + ksiz + vsiz;
    if(dmode == DP_DCAT) nsiz += head[DP_RHIVSIZ];
    if(nsiz <= rsiz){
      /* ¸ΰ˾ */
      if(!dprecover(depot, off, head, vbuf, vsiz, dmode == DP_DCAT)){
        depot->fatal = TRUE;
        return FALSE;
      }
    } else {
      /* ΰ */
      tval = NULL;
      if(dmode == DP_DCAT){
        if(!(tval = dprecval(depot, off, head, 0, -1))){
          depot->fatal = TRUE;
          return FALSE;
        }
        if(!(tval = realloc(tval, head[DP_RHIVSIZ] + vsiz))){
          free(tval);
          dpecode = DP_EALLOC;
          depot->fatal = TRUE;
          return FALSE;
        }
        memcpy(tval + head[DP_RHIVSIZ], vbuf, vsiz);
        vsiz += head[DP_RHIVSIZ];
        vbuf = tval;
      }
      /* ¸ΰκ */
      if(!dprecdelete(depot, off, head)){
        free(tval);
        depot->fatal = TRUE;
        return FALSE;
      }
      /* ΰ˽񤭹 */
      if((newoff = dprecappend(depot, kbuf, ksiz, vbuf, vsiz,
                               hash, head[DP_RHILEFT], head[DP_RHIRIGHT])) == -1){
        free(tval);
        depot->fatal = TRUE;
        return FALSE;
      }
      free(tval);
    }
    if(fdel) depot->rnum++;
    break;
  default:
    /* 񤭹 */
    if((newoff = dprecappend(depot, kbuf, ksiz, vbuf, vsiz, hash, 0, 0)) == -1){
      depot->fatal = TRUE;
      return FALSE;
    }
    depot->rnum++;
    break;
  }
  /* ι */
  if(newoff > 0){
    if(entoff > 0){
      if(!dpseekwritenum(depot->fd, entoff, newoff)){
        depot->fatal = TRUE;
        return FALSE;
      }
    } else {
      depot->bucket[bi] = newoff;
    }
  }
  return TRUE;
}


/* ǡ١쥳ɤ롣 */
int dpout(DEPOT *depot, const char *kbuf, int ksiz){
  int head[DP_RHNUM], hash, bi, off, entoff;
  assert(depot && kbuf);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  if(!depot->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* 쥳ɤθ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  hash = dpsecondhash(kbuf, ksiz);
  switch(dprecsearch(depot, kbuf, ksiz, hash, &bi, &off, &entoff, head, FALSE)){
  case -1:
    /* 顼 */
    depot->fatal = TRUE;
    return FALSE;
  case 0:
    /* ȯ */
    break;
  default:
    /* ʤ */
    dpecode = DP_ENOITEM;
    return FALSE;
  }
  /* ե饰ΩƤ */
  if(!dprecdelete(depot, off, head)){
    depot->fatal = TRUE;
    return FALSE;
  }
  depot->rnum--;
  return TRUE;
}


/* ǡ١쥳ɤͤФ롣 */
char *dpget(DEPOT *depot, const char *kbuf, int ksiz, int start, int max, int *sp){
  int head[DP_RHNUM], hash, bi, off, entoff;
  char *vbuf;
  assert(depot && kbuf && start >= 0);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return NULL;
  }
  /* 쥳ɤθ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  hash = dpsecondhash(kbuf, ksiz);
  switch(dprecsearch(depot, kbuf, ksiz, hash, &bi, &off, &entoff, head, FALSE)){
  case -1:
    /* 顼 */
    depot->fatal = TRUE;
    return NULL;
  case 0:
    /* ȯ */
    break;
  default:
    /* ʤ */
    dpecode = DP_ENOITEM;
    return NULL;
  }
  /* ͤμ¥ǡμ */
  if(start > head[DP_RHIVSIZ]){
    dpecode = DP_ENOITEM;
    return NULL;
  }
  if(!(vbuf = dprecval(depot, off, head, start, max))){
    depot->fatal = TRUE;
    return NULL;
  }
  if(sp){
    if(max < 0){
      *sp = head[DP_RHIVSIZ];
    } else {
      *sp = max < head[DP_RHIVSIZ] ? max : head[DP_RHIVSIZ];
    }
  }
  return vbuf;
}


/* ǡ١ˤ쥳ɤͤΥ롣 */
int dpvsiz(DEPOT *depot, const char *kbuf, int ksiz){
  int head[DP_RHNUM], hash, bi, off, entoff;
  assert(depot && kbuf);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return -1;
  }
  /* 쥳ɤθ */
  if(ksiz < 0) ksiz = strlen(kbuf);
  hash = dpsecondhash(kbuf, ksiz);
  switch(dprecsearch(depot, kbuf, ksiz, hash, &bi, &off, &entoff, head, FALSE)){
  case -1:
    /* 顼 */
    depot->fatal = TRUE;
    return -1;
  case 0:
    /* ȯ */
    break;
  default:
    /* ʤ */
    dpecode = DP_ENOITEM;
    return -1;
  }
  return head[DP_RHIVSIZ];
}


/* ǡ١ΥΥƥ졼롣 */
int dpiterinit(DEPOT *depot){
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  /* եåȤΥꥢ */
  depot->ioff = 0;
  return TRUE;
}


/* ǡ١ΥΥƥ졼鼡ΥФ */
char *dpiternext(DEPOT *depot, int *sp){
  int off, head[DP_RHNUM];
  char *val;
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return NULL;
  }
  /* ե饰ΤʤΥ쥳ɤФ */
  off = DP_HEADSIZ + depot->bnum * sizeof(int);
  off = off > depot->ioff ? off : depot->ioff;
  while(off < depot->fsiz){
    if(!dprechead(depot, off, head)){
      depot->fatal = TRUE;
      return NULL;
    }
    dpdbgouthead("dpiternext", head);
    if(head[DP_RHIFLAGS] & DP_RECFDEL){
      off += dprecsize(head);
    } else {
      if(!(val = dpreckey(depot, off, head))){
        depot->fatal = TRUE;
        return NULL;
      }
      depot->ioff = off + dprecsize(head);
      if(sp) *sp = head[DP_RHIKSIZ];
      return val;
    }
  }
  dpecode = DP_ENOITEM;
  return NULL;
}


/* ǡ١ƤǥƱ롣 */
int dpsync(DEPOT *depot){
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  if(!depot->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* إåȥХåƱ */
  *((int *)(depot->map + DP_FSIZOFF)) = depot->fsiz;
  *((int *)(depot->map + DP_RNUMOFF)) = depot->rnum;
  if(msync(depot->map, depot->msiz, MS_SYNC) == -1){
    dpecode = DP_EMAP;
    depot->fatal = TRUE;
    return FALSE;
  }
  /* 쥳Ʊ */
  if(fsync(depot->fd) == -1){
    dpecode = DP_ESYNC;
    depot->fatal = TRUE;
    return FALSE;
  }
  return TRUE;
}


/* ǡ١Υ饤Ȥꤹ롣 */
int dpsetalign(DEPOT *depot, int asiz, int aunit){
  assert(depot && asiz > 0 && aunit >= 0);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  if(!depot->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* 饤Ȥ */
  depot->asiz = asiz;
  depot->aunit = aunit;
  return TRUE;
}


/* ǡ١Ŭ롣 */
int dpoptimize(DEPOT *depot, int bnum){
  DEPOT *tdepot;
  char *name;
  int i, err, off, head[DP_RHNUM], ksizs[DP_OPTRUNIT], vsizs[DP_OPTRUNIT], unum, fd, msiz;
  char *kbufs[DP_OPTRUNIT], *vbufs[DP_OPTRUNIT], *map;
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return FALSE;
  }
  if(!depot->wmode){
    dpecode = DP_EMODE;
    return FALSE;
  }
  /* ǡ١κ */
  if(!(name = malloc(strlen(depot->name) + strlen(DP_OPTFSUF) + 1))){
    dpecode = DP_EALLOC;
    depot->fatal = FALSE;
    return FALSE;
  }
  sprintf(name, "%s%s", depot->name, DP_OPTFSUF);
  bnum = dpgetprime(bnum > 0 ? bnum : (int)(depot->rnum * (1.0 / DP_OPTBLOAD)) + 1);
  if(!(tdepot = dpopen(name, DP_OWRITER | DP_OCREAT | DP_OTRUNC, bnum))){
    free(name);
    depot->fatal = TRUE;
    return FALSE;
  }
  free(name);
  tdepot->asiz = depot->asiz;
  tdepot->aunit = depot->aunit;
  /* 쥳ɤΰܹ */
  err = FALSE;
  off = DP_HEADSIZ + depot->bnum * sizeof(int);
  unum = 0;
  while(off < depot->fsiz){
    if(!dprechead(depot, off, head)){
      err = TRUE;
      break;
    }
    dpdbgouthead("dpoptimize", head);
    if(!(head[DP_RHIFLAGS] & DP_RECFDEL)){
      kbufs[unum] = dpreckey(depot, off, head);
      ksizs[unum] = head[DP_RHIKSIZ];
      vbufs[unum] = dprecval(depot, off, head, 0, -1);
      vsizs[unum] = head[DP_RHIVSIZ];
      unum++;
      if(unum >= DP_OPTRUNIT){
        for(i = 0; i < unum; i++){
          if(kbufs[i] && vbufs[i]){
            if(!dpput(tdepot, kbufs[i], ksizs[i], vbufs[i], vsizs[i], DP_DOVER)) err = TRUE;
          } else {
            err = TRUE;
          }
          free(kbufs[i]);
          free(vbufs[i]);
          if(err) break;
        }
        unum = 0;
      }
    }
    off += dprecsize(head);
    if(err) break;
  }
  for(i = 0; i < unum; i++){
    if(kbufs[i] && vbufs[i]){
      if(!dpput(tdepot, kbufs[i], ksizs[i], vbufs[i], vsizs[i], DP_DOVER)) err = TRUE;
    } else {
      err = TRUE;
    }
    free(kbufs[i]);
    free(vbufs[i]);
    if(err) break;
  }
  if(err){
    unlink(tdepot->name);
    dpclose(tdepot);
    depot->fatal = TRUE;
    return FALSE;
  }
  /* ϥɥξ괹 */
  fd = depot->fd;
  map = depot->map;
  msiz = depot->msiz;
  depot->fd = tdepot->fd;
  depot->fsiz = tdepot->fsiz;
  depot->map = tdepot->map;
  depot->msiz = tdepot->msiz;
  depot->bucket = tdepot->bucket;
  depot->bnum = tdepot->bnum;
  depot->rnum = tdepot->rnum;
  depot->ioff = 0;
  depot->asiz = tdepot->asiz;
  depot->aunit = tdepot->aunit;
  if(rename(tdepot->name, depot->name) == -1){
    unlink(tdepot->name);
    dpecode = DP_ERENAME;
    err = TRUE;
  }
  /* ץ꥽β */
  free(tdepot->name);
  free(tdepot);
  if(munmap(map, msiz) == -1){
    dpecode = DP_EMAP;
    err = TRUE;
  }
  if(close(fd)){
    dpecode = DP_ECLOSE;
    err = TRUE;
  }
  if(err){
    depot->fatal = TRUE;
    return FALSE;
  }
  return TRUE;
}


/* ǡ١֤̾ */
char *dpname(DEPOT *depot){
  char *name;
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return NULL;
  }
  /* ̾ʣ */
  if(!(name = strdup(depot->name))){
    dpecode = DP_EALLOC;
    depot->fatal = TRUE;
    return NULL;
  }
  return name;
}


/* ǡ١Υե륵֤ */
int dpfsiz(DEPOT *depot){
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return -1;
  }
  return depot->fsiz;
}


/* ǡ١ΥХåǿ֤ */
int dpbnum(DEPOT *depot){
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return -1;
  }
  return depot->bnum;
}


/* ǡ١ΥХåǿ֤ */
int dpbusenum(DEPOT *depot){
  int i, hits;
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return -1;
  }
  /* Ψθ */
  hits = 0;
  for(i = 0; i < depot->bnum; i++){
    if(depot->bucket[i]) hits++;
  }
  return hits;
}


/* ǡ١Υ쥳ɿ֤ */
int dprnum(DEPOT *depot){
  assert(depot);
  /* ϥɥθ */
  if(depot->fatal){
    dpecode = DP_EFATAL;
    return -1;
  }
  return depot->rnum;
}


/* ǡ١Ѥϥåؿ */
int dpinnerhash(const char *kbuf, int ksiz){
  assert(kbuf);
  if(ksiz < 0) ksiz = strlen(kbuf);
  return dpfirsthash(kbuf, ksiz);
}


/* ǡ١ѤϥåؿΩϥåؿ */
int dpouterhash(const char *kbuf, int ksiz){
  assert(kbuf);
  if(ksiz < 0) ksiz = strlen(kbuf);
  return dpthirdhash(kbuf, ksiz);
}


/* ʾǿФ */
int dpprimenum(int num){
  assert(num > 0);
  return dpgetprime(num);
}



/*************************************************************************************************
 * ֥
 *************************************************************************************************/


/* եå롣
   fd оݤΥեǥץꤹ롣
   ex ʤ¾å򡢵ʤ鶦ͭåԤ
   ͤʤ鿿顼ʤ鵶Ǥ롣
   ¾ΥץƱե˥åƤ硢줬ޤǥ֥å롣 */
static int dplock(int fd, int ex){
  struct flock l;
  assert(fd >= 0);
  memset(&l, 0, sizeof(struct flock));
  l.l_type = ex ? F_WRLCK : F_RDLCK;
  l.l_whence = SEEK_SET;
  l.l_start = 0;
  l.l_len = 0;
  l.l_pid = 0;
  while(fcntl(fd, F_SETLKW, &l) == -1){
    if(errno != EINTR){
      dpecode = DP_ELOCK;
      return FALSE;
    }
  }
  return TRUE;
}


/* ե˥ХåեΥǡƽ񤭽Ф
   fd оݤΥեǥץꤹ롣
   buf Ͻ񤭽ФХåեꤹ롣
   size Ͻ񤭽ФХȿꤹ롣
   ͤʤ񤭽ФХȿ顼ʤ -1 Ǥ롣 */
static int dpwrite(int fd, const void *buf, int size){
  char *lbuf;
  int rv, wb;
  assert(fd >= 0 && buf && size >= 0);
  lbuf = (char *)buf;
  rv = 0;
  do {
    wb = write(fd, lbuf, size);
    switch(wb){
    case -1: if(errno != EINTR) return -1;
    case 0: break;
    default:
      lbuf += wb;
      size -= wb;
      rv += wb;
      break;
    }
  } while(size > 0);
  return rv;
}


/* ե֤˥ХåեΥǡƽ񤭽Ф
   fd оݤΥեǥץꤹ롣
   off Ͻ񤭽Ф֤ꤹ뤬ξϥեꤵ롣
   buf Ͻ񤭽ФХåեꤹ롣
   size Ͻ񤭽ФХȿꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dpseekwrite(int fd, int off, const void *buf, int size){
  char *lbuf;
  assert(fd >= 0 && buf && size >= 0);
  if(size < 1) return TRUE;
  lbuf = (char *)buf;
  if(off < 0){
    if(lseek(fd, 0, SEEK_END) == -1){
      dpecode = DP_ESEEK;
      return FALSE;
    }
  } else {
    if(lseek(fd, off, SEEK_SET) != off){
      dpecode = DP_ESEEK;
      return FALSE;
    }
  }
  if(dpwrite(fd, lbuf, size) != size){
    dpecode = DP_EWRITE;
    return FALSE;
  }
  return TRUE;
}


/* ե֤intοͤ񤭽Ф
   fd оݤΥեǥץꤹ롣
   off Ͻ񤭽Ф֤ꤹ뤬ξϥեꤵ롣
   num Ͻ񤭽Фͤꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dpseekwritenum(int fd, int off, int num){
  assert(fd >= 0);
  return dpseekwrite(fd, off, &num, sizeof(int));
}


/* ե뤫ХåեΥΥǡɤ߹ࡣ
   fd оݤΥեǥץꤹ롣
   buf ϥǡ¸Хåեꤹ롣
   size ɤ߹Хȿꤹ롣
   ͤʤɤ߹Хȿ顼ʤ -1 Ǥ롣
   ͤ size 꾮֤ͤäϥե뽪ụ̈롣 */
static int dpread(int fd, void *buf, int size){
  char *lbuf;
  int i, bs;
  assert(fd >= 0 && buf && size >= 0);
  lbuf = buf;
  for(i = 0; i < size && (bs = read(fd, lbuf + i, size - i)) != 0; i += bs){
    if(bs == -1 && errno != EINTR) return -1;
  }
  return i;
}


/* ե֤ХåեΥΥǡɤ߹ࡣ
   fd оݤΥեǥץꤹ롣
   off ɤ߹֤ꤹ롣
   buf ϥǡ¸Хåեꤹ롣
   size ɤ߹Хȿꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dpseekread(int fd, int off, void *buf, int size){
  char *lbuf;
  assert(fd >= 0 && off >= 0 && buf && size >= 0);
  lbuf = (char *)buf;
  if(lseek(fd, off, SEEK_SET) != off){
    dpecode = DP_ESEEK;
    return FALSE;
  }
  if(dpread(fd, lbuf, size) != size){
    dpecode = DP_EREAD;
    return FALSE;
  }
  return TRUE;
}


/* ե˥եޥåȤ˴ŤʸϤ롣
   fd оݤΥեǥץꤹ롣
   format ϥեޥåȤꤹ롣Ѵʸ %s %d %s %% ѤǤ롣
   裳ʹߤϥեޥåȤˤäƲѤȤʤ롣
   ͤʤ񤭽Фʸ顼ʤ鵶Ǥ롣 */
static int dpprintf(int fd, const char *format, ...){
  va_list ap;
  char numstr[DP_NUMBUFSIZ], *s;
  int len, wb;
  assert(fd >= 0 && format);
  va_start(ap, format);
  len = 0;
  while(*format != '\0'){
    if(*format == '%'){
      format++;
      switch(*format){
      case 's':
        s = va_arg(ap, char *);
        wb = dpwrite(fd, s, strlen(s));
        break;
      case 'd':
        sprintf(numstr, "%d", va_arg(ap, int));
        wb = dpwrite(fd, numstr, strlen(numstr));
        break;
      case 'x':
        sprintf(numstr, "%x", va_arg(ap, int));
        wb = dpwrite(fd, numstr, strlen(numstr));
        break;
      case '%':
        wb = dpwrite(fd, "%", 1);
        break;
      case '\0':
        va_end(ap);
        return -1;
      default:
        wb = 0;
        break;
      }
      if(wb == -1){
        va_end(ap);
        return -1;
      }
      len += wb;
    } else {
      if(dpwrite(fd, format, 1) == -1){
        va_end(ap);
        return -1;
      }
      len++;
    }
    format++;
  }
  va_end(ap);
  return len;
}


/* ǥХå˥ǡ١ϥɥξϤ롣
   fname ǤդƬꤹ롣
   depot ϥǡ١ϥɥꤹ롣 */
static void dpdbgoutdepot(const char *fname, DEPOT *depot){
  assert(fname && depot);
  if(dpdbgfd < 0) return;
  dpprintf(dpdbgfd, "%s: depot: name=%s wmode=%d fd=%d fsiz=%d map=0x%x msiz=%d "
           "bucket=0x%x bnum=%d rnum=%d fatal=%d asiz=%d, aunit=%d\n",
           fname, depot->name, depot->wmode, depot->fd, depot->fsiz, depot->map, depot->msiz,
           depot->bucket, depot->bnum, depot->rnum, depot->fatal, depot->asiz, depot->aunit);
}


/* ǥХå˥쥳ɥإåξϤ롣
   fname ǤդƬꤹ롣
   head ϥ쥳ɤΥإåǼꤹ롣 */
static void dpdbgouthead(const char *fname, int *head){
  assert(fname && head);
  if(dpdbgfd < 0) return;
  dpprintf(dpdbgfd, "%s: head: flags=%d hash=%d ksiz=%d vsiz=%d psiz=%d "
           "left=%d right=%d [siz=%d]\n",
           fname, head[DP_RHIFLAGS], head[DP_RHIHASH], head[DP_RHIKSIZ], head[DP_RHIVSIZ],
           head[DP_RHIPSIZ], head[DP_RHILEFT], head[DP_RHIRIGHT], dprecsize(head));
}


/* 裱ϥå֤ͤ
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   ͤĴΥϥåͤǤꡢХåǿǾ;򻻽ФƤѤ٤Ǥ롣 */
static int dpfirsthash(const char *kbuf, int ksiz){
  const unsigned char *p;
  unsigned int sum;
  int i;
  assert(kbuf && ksiz >= 0);
  p = (const unsigned char *)kbuf;
  sum = 19780211 + ksiz;
  for(i = 0; i < ksiz; i++){
    sum = sum * 31 + p[i];
  }
  return ((sum * 787767623) + 2003) & 0x7FFFFFFF;
}


/* 裲ϥå֤ͤ
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   ͤĴΥϥåͤǤꡢХåǿǾ;򻻽ФƤѤ٤Ǥ롣 */
static int dpsecondhash(const char *kbuf, int ksiz){
  const unsigned char *p;
  unsigned int sum;
  int i;
  assert(kbuf && ksiz >= 0);
  p = (const unsigned char *)kbuf;
  sum = 11;
  for(i = ksiz - 1; i >= 0; i--){
    sum = sum * 37 + p[i];
  }
  return ((sum * 7340009) + ksiz) & 0x7FFFFFFF;
}


/* 裳ϥå֤ͤ 
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   ͤĴΥϥåͤǤꡢХåǿǾ;򻻽ФƤѤ٤Ǥ롣 */
static int dpthirdhash(const char *kbuf, int ksiz){
  const unsigned char *p;
  unsigned int sum;
  int i;
  assert(kbuf && ksiz >= 0);
  p = (const unsigned char *)kbuf;
  sum = 751;
  for(i = ksiz - 1; i >= 0; i--){
    sum = sum * 29 + p[i];
  }
  return (sum * 10123411) & 0x7FFFFFFF;
}


/* ʾǿФ
   num Ŭοꤹ롣
   ͤϡꤷƱ礭Ĥʤ٤ͤǤ롣 */
static int dpgetprime(int num){
  int primes[] = {
    1, 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 43, 47, 53, 59, 61, 71, 79, 83,
    89, 103, 109, 113, 127, 139, 157, 173, 191, 199, 223, 239, 251, 283, 317, 349,
    383, 409, 443, 479, 509, 571, 631, 701, 761, 829, 887, 953, 1021, 1151, 1279,
    1399, 1531, 1663, 1789, 1913, 2039, 2297, 2557, 2803, 3067, 3323, 3583, 3833,
    4093, 4603, 5119, 5623, 6143, 6653, 7159, 7673, 8191, 9209, 10223, 11261,
    12281, 13309, 14327, 15359, 16381, 18427, 20479, 22511, 24571, 26597, 28669,
    30713, 32749, 36857, 40949, 45053, 49139, 53239, 57331, 61417, 65521, 73727,
    81919, 90107, 98299, 106487, 114679, 122869, 131071, 147451, 163819, 180221,
    196597, 212987, 229373, 245759, 262139, 294911, 327673, 360439, 393209, 425977,
    458747, 491503, 524287, 589811, 655357, 720887, 786431, 851957, 917503, 982981,
    1048573, 1179641, 1310719, 1441771, 1572853, 1703903, 1835003, 1966079,
    2097143, 2359267, 2621431, 2883577, 3145721, 3407857, 3670013, 3932153,
    4194301, 4718579, 5242877, 5767129, 6291449, 6815741, 7340009, 7864301,
    8388593, 9437179, 10485751, 11534329, 12582893, 13631477, 14680063, 15728611,
    16777213, 18874367, 20971507, 23068667, 25165813, 27262931, 29360087, 31457269,
    33554393, 37748717, 41943023, 46137319, 50331599, 54525917, 58720253, 62914549,
    67108859, 75497467, 83886053, 92274671, 100663291, 109051903, 117440509,
    125829103, 134217689, 150994939, 167772107, 184549373, 201326557, 218103799,
    234881011, 251658227, 268435399, 301989881, 335544301, 369098707, 402653171,
    436207613, 469762043, 503316469, 536870909, 603979769, 671088637, 738197503,
    805306357, 872415211, 939524087, 1006632947, 1073741789, 1207959503,
    1342177237, 1476394991, 1610612711, 1744830457, 1879048183, 2013265907, -1
  };
  int i;
  assert(num > 0);
  for(i = 0; primes[i] > 0; i++){
    if(num <= primes[i]) return primes[i];
  }
  return primes[i-1];
}


/* 쥳ɤΥѥǥ󥰤Υ֤
   vsiz ϥ쥳ɤͤΥꤹ롣
   ͤϥѥǥ󥰤ΥǤ롣 */
static int dppadsize(DEPOT *depot, int vsiz){
  int abase;
  assert(depot && vsiz >= 0);
  if(depot->asiz < 1 || depot->aunit < 1) return 0;
  abase = vsiz / depot->aunit + 1;
  return ((abase * depot->aunit) - vsiz) % (depot->asiz * abase);
}


/* 쥳ɤΥեǤΥ֤
   head ϥ쥳ɤΥإåǼꤹ롣
   ͤϡΥ쥳ɤΤ˥ե˳ݤƤϢΰΥǤ롣 */
static int dprecsize(int *head){
  assert(head);
  return DP_RHNUM * sizeof(int) + head[DP_RHIKSIZ] + head[DP_RHIVSIZ] + head[DP_RHIPSIZ];
}


/* 쥳ɤΥإåɤ߹ࡣ
   depot ϥǡ١ϥɥꤹ롣
   off ϥ쥳ɤΥեåȤꤹ롣
   head ΰˤϥ쥳ɤΥإå󤬳Ǽ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dprechead(DEPOT *depot, int off, int *head){
  assert(depot && off >= 0 && head);
  if(off > depot->fsiz){
    dpecode = DP_EBROKEN;
    return FALSE;
  }
  if(!dpseekread(depot->fd, off, head, DP_RHNUM * sizeof(int))) return FALSE;
  if(head[DP_RHIKSIZ] < 0 || head[DP_RHIVSIZ] < 0 || DP_RHIPSIZ < 0 ||
     head[DP_RHILEFT] < 0 || head[DP_RHIRIGHT] < 0){
    dpecode = DP_EBROKEN;
    return FALSE;
  }
  return TRUE;
}


/* 쥳ɤΥμ¥ǡɤ߹ࡣ
   depot ϥǡ١ϥɥꤹ롣
   off ϥ쥳ɤΥեåȤꤹ롣
   head ϥ쥳ɤΥإåǼꤹ롣
   ͤϡǼʸ򥢥ȤΰΥݥ󥿤顼ʤ NULL Ǥ롣
   ͤΰϡ줿ΰ꣱Х¿ݤƽüʸ֤롣 */
static char *dpreckey(DEPOT *depot, int off, int *head){
  char *kbuf;
  int ksiz;
  assert(depot && off >= 0);
  ksiz = head[DP_RHIKSIZ];
  if(!(kbuf = malloc(ksiz + 1))){
    dpecode = DP_EALLOC;
    return NULL;
  }
  if(!dpseekread(depot->fd, off + DP_RHNUM * sizeof(int), kbuf, ksiz)){
    free(kbuf);
    return NULL;
  }
  kbuf[ksiz] = '\0';
  return kbuf;
}


/* 쥳ɤΥμ¥ǡɤ߹ࡣ
   depot ϥǡ١ϥɥꤹ롣
   off ϥ쥳ɤΥեåȤꤹ롣
   head ϥ쥳ɤΥإåǼꤹ롣
   start ϼФͤƬΥեåȤꤹ롣
   max ϼФͤκĹꤹ롣
   ͤϡǼʸ򥢥ȤΰΥݥ󥿤顼ʤ NULL Ǥ롣
   ͤΰϡ줿ΰ꣱Х¿ݤƽüʸ֤롣 */
static char *dprecval(DEPOT *depot, int off, int *head, int start, int max){
  char *vbuf;
  int vsiz;
  assert(depot && off >= 0 && start >= 0);
  head[DP_RHIVSIZ] -= start;
  if(max < 0){
    vsiz = head[DP_RHIVSIZ];
  } else {
    vsiz = max < head[DP_RHIVSIZ] ? max : head[DP_RHIVSIZ];
  }
  if(!(vbuf = malloc(vsiz + 1))){
    dpecode = DP_EALLOC;
    return NULL;
  }
  if(!dpseekread(depot->fd, off + DP_RHNUM * sizeof(int) + head[DP_RHIKSIZ] + start, vbuf, vsiz)){
    free(vbuf);
    return NULL;
  }
  vbuf[vsiz] = '\0';
  return vbuf;
}


/* μΥǡƱӤԤ
   abuf ԤΥǡΰΥݥ󥿤ꤹ롣
   asiz  abuf ΰΥꤹ롣
   bbuf ϸԤΥǡΰΥݥ󥿤ꤹ롣
   bsiz  bbuf ΰΥꤹ롣
   ͤϡ¾Ʊʤ 0 Ԥ礭ǤʤǤ롣 */
static int dpkeycmp(const char *abuf, int asiz, const char *bbuf, int bsiz){
  assert(abuf && asiz >= 0 && bbuf && bsiz >= 0);
  if(asiz > bsiz) return 1;
  if(asiz < bsiz) return -1;
  return memcmp(abuf, bbuf, asiz);
}


/* 쥳ɤǡ١鸡롣
   depot ϥǡ١ϥɥꤹ롣
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   hash ϥ裲ϥåͤꤹ롣
   bip λˤϳ쥳ɤΥϥåź롣
   offp λˤϥϥåǺǸˤɤä쥳ɤΥեåȤ
   ϥå󤬶ʤ -1 롣
   entp λˤϺǸѤʥإåκˤΥեåȤ뤫
   ϥå󤬶ʤ -1 롣
   head ΰˤϺǸˤɤä쥳ɤΥإå󤬳Ǽ롣
   delhit ʤե饰ĤǤǤ⳺Ȥߤʤʤߤʤʤ
   ͤϡ׳Ǥä 0 ʤä 1 顼ʤ -1 Ǥ롣
   פˤ offp λ褬ǤΥեåȤˤʤꡢΥإå head ˤʤ롣
   פʤäˤϡɲä٤Ǥؤ entp λȤʤ롣 */
static int dprecsearch(DEPOT *depot, const char *kbuf, int ksiz, int hash,
                       int *bip, int *offp, int *entp, int *head, int delhit){
  int off, entoff, thash, kcmp;
  char *tkey;
  assert(depot && kbuf && ksiz >= 0 && hash >= 0 && bip && offp && entp && head);
  /* ϥåμ */
  *bip = dpfirsthash(kbuf, ksiz) % depot->bnum;
  off = depot->bucket[*bip];
  *offp = -1;
  *entp = -1;
  /* ϥåõ */
  entoff = -1;
  while(off != 0){
    /* إåɤ߹ */
    if(!dprechead(depot, off, head)) return -1;
    /* ǥХå */
    dpdbgouthead("dprecsearch", head);
    /* 裲ϥåͤˤ */
    thash = head[DP_RHIHASH];
    if(hash > thash){
      entoff = off + DP_RHILEFT * sizeof(int);
      off = head[DP_RHILEFT];
    } else if(hash < thash){
      entoff = off + DP_RHIRIGHT * sizeof(int);
      off = head[DP_RHIRIGHT];
    } else {
      /* ¥ǡˤ */
      if(!(tkey = dpreckey(depot, off, head))) return -1;
      kcmp = dpkeycmp(kbuf, ksiz, tkey, head[DP_RHIKSIZ]);
      free(tkey);
      if(kcmp > 0){
        entoff = off + DP_RHILEFT * sizeof(int);
        off = head[DP_RHILEFT];
      } else if(kcmp < 0){
        entoff = off + DP_RHIRIGHT * sizeof(int);
        off = head[DP_RHIRIGHT];
      } else {
        /* ׽ */
        if(!delhit && (head[DP_RHIFLAGS] & DP_RECFDEL)){
          entoff = off + DP_RHILEFT * sizeof(int);
          off = head[DP_RHILEFT];
        } else {
          *offp = off;
          *entp = entoff;
          return 0;
        }
      }
    }
  }
  *offp = off;
  *entp = entoff;
  return 1;
}


/* 쥳ɤե˽񤭹ࡣ
   depot ϥǡ١ϥɥꤹ롣
   kbuf ϥΥǡΰΥݥ󥿤ꤹ롣
   ksiz ϥΥǡΰΥХȿǻꤹ롣
   vbuf ͤΥǡΰΥݥ󥿤ꤹ롣
   vsiz ͤΥǡΰΥХȿǻꤹ롣
   hash ϥ裲ϥåͤꤹ롣
   left ϺλҤΥեåȤꤹ롣
   right ϱλҤΥեåȤꤹ롣
   ͤϡʤ񤭹쥳ɤΥեåȡ顼ʤ -1 Ǥ롣 */
static int dprecappend(DEPOT *depot, const char *kbuf, int ksiz, const char *vbuf, int vsiz,
                       int hash, int left, int right){
  int head[DP_RHNUM], psiz, off, hoff, koff, voff, c;
  assert(depot && kbuf && ksiz >= 0 && vbuf && vsiz >= 0);
  /* إå */
  psiz = dppadsize(depot, vsiz);
  head[DP_RHIFLAGS] = 0;
  head[DP_RHIHASH] = hash;
  head[DP_RHIKSIZ] = ksiz;
  head[DP_RHIVSIZ] = vsiz;
  head[DP_RHIPSIZ] = psiz;
  head[DP_RHILEFT] = left;
  head[DP_RHIRIGHT] = right;
  /* եåȷ׻ */
  off = depot->fsiz;
  hoff = off;
  koff = hoff + sizeof(head);
  voff = koff + ksiz;
  /* ǥХå */
  dpdbgouthead("dprecappend", head);
  /* 񤭹 */
  c = 0;
  if(!dpseekwrite(depot->fd, hoff, head, sizeof(head)) ||
     !dpseekwrite(depot->fd, koff, kbuf, ksiz) ||
     !dpseekwrite(depot->fd, voff, vbuf, vsiz) ||
     (psiz > 0 && !dpseekwrite(depot->fd, voff + vsiz + psiz - 1, &c, 1))) return -1;
  depot->fsiz += sizeof(head) + vsiz + ksiz + psiz;
  return off;
}


/* 쥳ɤ񤭤롣
   depot ϥǡ١ϥɥꤹ롣
   off ϥ쥳ɤΥեåȤꤹ롣
   head ϥ쥳ɤΥإåǼꤹ롣
   vbuf ͤΥǡΰΥݥ󥿤ꤹ롣
   vsiz ͤΥǡΰΥХȿǻꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dprecover(DEPOT *depot, int off, int *head, const char *vbuf, int vsiz, int cat){
  int hsiz, hoff, voff;
  assert(depot && off >= 0 && head && vbuf && vsiz >= 0);
  if(cat){
    /* إå */
    head[DP_RHIFLAGS] = 0;
    head[DP_RHIPSIZ] -= vsiz;
    head[DP_RHIVSIZ] += vsiz;
    /* եåȷ׻ */
    hsiz = DP_RHNUM * sizeof(int);
    hoff = off;
    voff = hoff + DP_RHNUM * sizeof(int) + head[DP_RHIKSIZ] + head[DP_RHIVSIZ] - vsiz;
  } else {
    /* إå */
    head[DP_RHIFLAGS] = 0;
    head[DP_RHIPSIZ] += head[DP_RHIVSIZ] - vsiz;
    head[DP_RHIVSIZ] = vsiz;
    /* եåȷ׻ */
    hsiz = DP_RHNUM * sizeof(int);
    hoff = off;
    voff = hoff + DP_RHNUM * sizeof(int) + head[DP_RHIKSIZ];
  }
  /* ǥХå */
  dpdbgouthead("dprecover", head);
  /* 񤭹 */
  if(!dpseekwrite(depot->fd, hoff, head, hsiz) ||
     !dpseekwrite(depot->fd, voff, vbuf, vsiz)) return FALSE;
  return TRUE;
}


/* 쥳ɤκե饰ΩƤ롣
   depot ϥǡ١ϥɥꤹ롣
   off ϥ쥳ɤΥեåȤꤹ롣
   head ϥ쥳ɤΥإåǼꤹ롣
   ͤʤ鿿顼ʤ鵶Ǥ롣 */
static int dprecdelete(DEPOT *depot, int off, int *head){
  assert(depot && off >= 0 && head);
  return dpseekwritenum(depot->fd, off + DP_RHIFLAGS * sizeof(int),
                        head[DP_RHIFLAGS] | DP_RECFDEL);
}



/* END OF FILE */
